
#include "parser.h"
#include "lexer.h"

#include <qstring.h>
#include <qstringlist.h>

using namespace CppSupport;

#define qDebug(a...)

#define ADVANCE(tk, descr) \
{ \
    Token token = lex->lookAhead( 0 ); \
    if( token != tk ){ \
        reportError( tr("expected %1 found %2").arg((descr)).arg(token.toString()), \
                     token ); \
        return false; \
    } \
    lex->nextToken(); \
}

#define MATCH(tk, descr) \
{ \
    Token token = lex->lookAhead( 0 ); \
    if( token != tk ){ \
        reportError( tr("expected %1 found %2").arg((descr)).arg(token.toString()), \
                     token ); \
        return false; \
    } \
}

Parser::Parser( Lexer* lexer )
    : lex( lexer )
{
    m_fileName = "<stdin>";

    m_dom = new QDomDocument( "Cpp" );
    m_translationUnit = m_dom->createElement( "translation-unit" );
    m_dom->appendChild( m_translationUnit );

    resetErrors();
}

Parser::~Parser()
{
    delete( m_dom );
    m_dom = 0;
}

void Parser::setFileName( const QString& fileName )
{
    m_translationUnit.setAttribute( "file-name", fileName );
    m_fileName = fileName;
}

void Parser::reportError( const QString& msg, const Token& token )
{
    int line=0, col=0;
    lex->getTokenPosition( token, &line, &col );

    qWarning( "%s:%d:%d: %s",
              m_fileName.latin1(),
              line,
              col,
              msg.latin1() );
    ++m_errors;
}

void Parser::reportWarning( const QString& msg, const Token& token )
{
    int line=0, col=0;
    lex->getTokenPosition( token, &line, &col );

    qWarning( "%s:%d:%d: %s",
              m_fileName.latin1(),
              line,
              col,
              msg.latin1() );
}

void Parser::syntaxError()
{
    reportError( "syntax error", lex->lookAhead(0) );
}

bool Parser::skipUntil( int token )
{
    while( !lex->lookAhead(0).isNull() ){
        if( lex->lookAhead(0) == token )
            return true;

        lex->nextToken();
    }
    return false;
}

bool Parser::skip( int l, int r )
{
    int count = 0;
    while( !lex->lookAhead(0).isNull() ){
        int tk = lex->lookAhead( 0 );

        if( tk == l )
            ++count;
        else if( tk == r )
            --count;

        if( count == 0 )
            return true;

        lex->nextToken();
    }
    return false;
}

// name : {'::'} name2 ('::' name2)*
bool Parser::parseName( QDomElement& name )
{
    name = m_dom->createElement( "nested-name" );

    if( lex->lookAhead(0) == Token_scope ){
        name.setAttribute( "global", true );
        lex->nextToken();
    }

    QDomElement name2;
    if( !parseName2(name2) ){
        return false;
    }
    name.appendChild( name2 );

    while( lex->lookAhead(0) == Token_scope ){
        lex->nextToken();

        if( !parseName2(name2) ){
            reportError( "expected a name", lex->lookAhead(0) );
            return false;
        }
        name.appendChild( name2 );
    }
    return true;
}


//  name2:
//    Identifier {template.args}
//  | '~' Identifier
//  | OPERATOR operator.name {template.args}
bool Parser::parseName2( QDomElement& name )
{
    bool isDestructor = false;

    if( lex->lookAhead(0) == Token_identifier ){
        name = m_dom->createElement( "name" );
        name.setAttribute( "id", lex->lookAhead(0).toString() );
        lex->nextToken();
    } else if( lex->lookAhead(0) == '~' && lex->lookAhead(1) == Token_identifier ){
        name = m_dom->createElement( "destructor" );
        lex->nextToken(); // skip ~
        name.setAttribute( "id", lex->lookAhead(0).toString() );
        lex->nextToken(); // skip classname
        isDestructor = true;
    } else if( lex->lookAhead(0) == Token_operator ){
        return parseOperatorFunctionId( name );
//         lex->nextToken(); // skip operator
//         if( !parseOperator(name) ){
//             reportError( "expected operator", lex->lookAhead(0) );
//             return false;
//         }
    } else
        return false;

    if( !isDestructor ){

        if( lex->lookAhead(0) == '<' ){
            lex->nextToken();

            // optional template arguments
            QDomElement templArgs;
            if( parseTemplateArgumentList(templArgs) ){
                name.appendChild( templArgs );
            }

            ADVANCE( '>', '>' );
        }
    }
    return true;
}


bool Parser::parseTranslationUnit()
{
    qDebug( "Parser::parseTranslationUnit()" );

    QDomElement decl;

    while( !lex->lookAhead(0).isNull() ){

        if( parseDefinition(decl) ){
            m_translationUnit.appendChild( decl );
        } else {
            // error recovery
            syntaxError();
            if( skipUntil(';') ){
                lex->nextToken(); // skip ;
            }
        }
    }

    return true;
}

bool Parser::parseDefinition( QDomElement& decl )
{
    qDebug( "Parser::parseDefinition()" );

    switch( lex->lookAhead(0) ){

    case ';':
        lex->nextToken();
        decl = m_dom->createElement( "null-declaration" );
        return true;

    case Token_extern:
        return parseLinkageSpecification( decl );

    case Token_namespace:
        return parseNamespace( decl );

    case Token_using:
        return parseUsing( decl );

    case Token_typedef:
        return parseTypedef( decl );

    case Token_asm:
        return parseAsmDefinition( decl );

    case Token_template:
    case Token_export:
        return parseTemplateDeclaration( decl );

    default:
        return parseDeclaration( decl );

    } // end switch
}


bool Parser::parseLinkageSpecification( QDomElement& spec )
{
    if( lex->lookAhead(0) != Token_extern ){
        return false;
    }
    lex->nextToken();

    spec = m_dom->createElement( "linkage-specification" );

    if( lex->lookAhead(0) == Token_string_literal ){
        spec.setAttribute( "type", lex->lookAhead(0).toString() );
        lex->nextToken(); // skip string literal
    }

    if( lex->lookAhead(0) == '{' ){
        QDomElement body;
        if( parseLinkageBody(body) ){
            spec.appendChild( body );
        }
    } else {
        QDomElement def;
        if( !parseDefinition(def) ){
            return false;
        }
        spec.appendChild( def );
    }

    return true;
}

bool Parser::parseLinkageBody( QDomElement& body )
{
    body = m_dom->createElement( "linkage-body" );

    if( lex->lookAhead(0) != '{' ){
        return false;
    }
    lex->nextToken();

    while( !lex->lookAhead(0).isNull() ){
        int tk = lex->lookAhead( 0 );

        if( tk == '}' )
            break;

        QDomElement def;
        if( !parseDefinition(def) ){
            // error recovery
            reportError( "expected a definition", lex->lookAhead(0) );
            if( skipUntil(';') ){
                lex->nextToken(); // skip ;
            }
        } else {
            body.appendChild( def );
        }
    }

    if( lex->lookAhead(0) != '}' ){
        // error recovery
        reportError( "expected }", lex->lookAhead(0) );
        skipUntil( '}' );
    }

    lex->nextToken(); // skip }

    return true;
}

bool Parser::parseNamespace( QDomElement& spec )
{
    if( lex->lookAhead(0) != Token_namespace ){
        return false;
    }
    lex->nextToken();

    spec = m_dom->createElement( "namespace-specification" );

    QString ide;
    if( lex->lookAhead(0) == Token_identifier ){
        ide = lex->lookAhead( 0 ).toString();
        spec.setAttribute( "id", ide );
        lex->nextToken();
    }

    if( lex->lookAhead(0) == '{' ){

        QDomElement body;
        if( parseLinkageBody(body) ){
            spec.appendChild( body );
        }

    }

    return true;
}

bool Parser::parseUsing( QDomElement& decl )
{
    if( lex->lookAhead(0) != Token_using ){
        return false;
    }
    lex->nextToken();

    if( lex->lookAhead(0) == Token_namespace ){
        return parseUsingDirective( decl );
    }

    decl = m_dom->createElement( "using-declaration" );

    QDomElement name;
    if( parseName(name) ){
        decl.appendChild( name );

        if( lex->lookAhead(0) != ';' ){
            // TODO error recovery
        }
        lex->nextToken(); // skip ;
        return true;
    }

    if( lex->lookAhead(0) == Token_typename ){
        decl.setAttribute( "typename", true );
        lex->nextToken();
    }

    if( !parseName(name) ){
        return false;
    }
    decl.appendChild( name );

    if( lex->lookAhead(0) != ';' ){
        return false;
    }
    lex->nextToken();

    return true;
}

bool Parser::parseUsingDirective( QDomElement& def )
{
    if( lex->lookAhead(0) != Token_namespace ){
        return false;
    }
    lex->nextToken();

    def = m_dom->createElement( "using-directive" );

    QDomElement name;
    if( !parseName(name) ){
        return false;
    }
    def.appendChild( name );

    if( lex->lookAhead(0) != ';' ){
        return false;
    }
    lex->nextToken();

    return true;
}


// ----------------------------------------------------------------------
// TODO: review
// ----------------------------------------------------------------------

bool Parser::parseOperatorFunctionId( QDomElement& id )
{
    if( lex->lookAhead(0) != Token_operator ){
        return false;
    }
    lex->nextToken();

    id = m_dom->createElement( "operator-function-id" );

    if( parseOperator(id) )
        return true;
    else {
        // parse cast operator
        id.setAttribute( "cast", true );
        QDomElement cv;
        if( parseCvQualify(cv) ){
            id.appendChild( cv );
        }

        QDomElement typeSpec;
        if( !parseSimpleTypeSpecifier(typeSpec) ){
            return false;
        }
        id.appendChild( typeSpec );

        if( parseCvQualify(cv) ){
            id.appendChild( cv );
        }

        QDomElement ptrOp;
        while( parsePtrOperator(ptrOp) ){
            id.appendChild( ptrOp );
        }
        return true;
    }

    return false;
}

bool Parser::parseTemplateArgumentList( QDomElement& argumentList )
{
    argumentList = m_dom->createElement( "template-argument-list" );

    QDomElement arg;
    if( !parseTemplateArgument(arg) )
        return false;

    argumentList.appendChild( arg );

    while( lex->lookAhead(0) == ',' ){
        lex->nextToken();

        if( parseTemplateArgument(arg) ){
            argumentList.appendChild( arg );
        } else
            break;
    }

    return true;
}

bool Parser::parseTypedef( QDomElement& def )
{
    // qDebug( "Parser::parseTypedef()" );
    if( lex->lookAhead(0) != Token_typedef ){
        return false;
    }
    lex->nextToken();

    def = m_dom->createElement( "typedef-definition" );

    QDomElement typeSpec;
    if( !parseTypeSpecifier(typeSpec) ){
        return false;
    }

    QDomElement declarator;
    if( !parseDeclarator(declarator) ){
        return false;
    }

    ADVANCE( ';', ';' );

    def.appendChild( typeSpec );
    def.appendChild( declarator );

    return true;
}

bool Parser::parseAsmDefinition( QDomElement& definition )
{
    ADVANCE( Token_asm, "asm" );
    ADVANCE( '(', '(' );

    MATCH( Token_string_literal, "string literal" );
    QString body = lex->lookAhead( 0 ).toString();
    lex->nextToken();

    ADVANCE( ')', ')' );
    ADVANCE( ';', ';' );

    definition = m_dom->createElement( "asm-definition" );
    QDomText e = m_dom->createTextNode( body );
    definition.appendChild( e );

    return true;
}

bool Parser::parseTemplateDeclaration( QDomElement& templ_decl )
{
    bool _export = false;
    if( lex->lookAhead(0) == Token_export ){
        _export = true;
        lex->nextToken();
    }

    if( lex->lookAhead(0) != Token_template ){
        if( _export ){
            ADVANCE( Token_template, "template" );
        } else
            return false;
    }

    ADVANCE( Token_template, "template" );

    if( lex->lookAhead(0) == '<' ){
        lex->nextToken();

        templ_decl = m_dom->createElement( "template-declaration" );
        templ_decl.setAttribute( "export", _export );

        QDomElement parameterList;
        if( parseTemplateParameterList(parameterList) ){
            templ_decl.appendChild( parameterList );
        }

        ADVANCE( '>', '>' );
    }

    QDomElement decl;
    if( !parseDefinition( decl ) ){
        reportError( "expected a declaration", lex->lookAhead( 0 ) );
        return false;
    }
    templ_decl.appendChild( decl );

    return true;
}

bool Parser::parseDeclaration( QDomElement& decl )
{
    return
        parseIntegralDeclaration( decl ) ||
        parseConstDeclaration( decl ) ||
        parseOtherDeclaration( decl );
}

bool Parser::parseDeclHead( QDomElement& head )
{
    head = m_dom->createElement( "declaration-head" );

    QDomElement spec;
    while( parseStorageClassSpecifier(spec) || parseFunctionSpecifier(spec) ){
        head.appendChild( spec );
    }

    QDomElement cv;
    if( parseCvQualify(cv) ){
        head.appendChild( cv );
    }

    return true;
}

bool Parser::parseIntegralDeclaration( QDomElement& decl )
{
    int index = lex->index();

    decl = m_dom->createElement( "integral-declaration" );

    QDomElement head;
    if( !parseIntegralDeclHead(head) ){
        lex->setIndex( index );
        return false;
    }
    decl.appendChild( head );

    int tk = lex->lookAhead( 0 );
    if( tk == ';' ){
        lex->nextToken();
    } else if( tk == ':' ){
        lex->nextToken();
        QDomElement expr;
        if( !parseConstantExpression(expr) ){
            lex->setIndex( index );
            return false;
        }
        decl.appendChild( expr );
        ADVANCE( ';', ';' );
    } else {
        QDomElement declarators;
        if( !parseInitDeclaratorList(declarators) ){
            lex->setIndex( index );
            return false;
        }
        decl.appendChild( declarators );
        if( lex->lookAhead(0) == ';' ){
            lex->nextToken();
        } else {
            QDomElement body;
            if( !parseFunctionBody(body) ){
                lex->setIndex( index );
                return false;
            }
            decl.appendChild( body );
        }
    }

    return true;
}

bool Parser::parseIntegralDeclHead( QDomElement& head )
{
    head = m_dom->createElement( "integral-declaration-head" );

    QDomElement declHead;
    if( !parseDeclHead(declHead) ){
        return false;
    }
    head.appendChild( declHead );

    QDomElement typeSpec;
    if( !parseTypeSpecifier(typeSpec) ){
        return false;
    }
    head.appendChild( typeSpec );

    QDomElement cv;
    if( !parseCvQualify(cv) ){
        return false;
    }
    head.appendChild( cv );


    return true;
}

bool Parser::parseOtherDeclaration( QDomElement& decl )
{
    decl = m_dom->createElement( "other-declaration" );

    if( lex->lookAhead(0) == Token_friend ){
        lex->nextToken();
        if( !parseDefinition(decl) ){
            reportError( "expected a declaration", lex->lookAhead(0) );
            return false;
        }
        decl.setAttribute( "friend", true );
    } else {

        QDomElement head;
        if( !parseDeclHead(head) ){
            return false;
        }
        decl.appendChild( head );

        QDomElement name;
        if( !parseNestedNameSpecifier(name) ){
            return false;
        }
        decl.appendChild( name );

        if( isConstructorDecl() ){
            QDomElement constr;
            if( !parseConstructorDeclaration(constr) ){
                return false;
            }
            decl.appendChild( constr );
        } else {

            QDomElement cv;
            if( parseCvQualify(cv) ){
                decl.appendChild( cv );
            }

            QDomElement declarators;
            if( !parseInitDeclaratorList(declarators) ){
                return false;
            }
            decl.appendChild( declarators );
        }

        if( lex->lookAhead(0) == ';' ){
            lex->nextToken();
        } else {
            QDomElement body;
            if( !parseFunctionBody(body) ){
                return false;
            }
            decl.appendChild( body );
        }
    }


    return true;
}

bool Parser::parseConstDeclaration( QDomElement& decl )
{
    decl = m_dom->createElement( "const-declaration" );

    int index = lex->index();

    QDomElement cv;
    if( parseCvQualify(cv) ){
        decl.appendChild( cv );
    }

    if( lex->lookAhead(0) == '*' ){
        decl.setAttribute( "star", true );
        lex->nextToken();
    }

    if( lex->lookAhead(0) != Token_identifier ){
        lex->setIndex( index );
        return false;
    }
    decl.setAttribute( "ide1", lex->lookAhead(0).toString() );
    lex->nextToken();

    if( lex->lookAhead(0) != '=' ){
        lex->setIndex( index );
        return false;
    }
    lex->nextToken();

    QDomElement expr;
    if( !parseConstantExpression(expr) ){
        lex->setIndex( index );
        return false;
    }
    decl.appendChild( expr );

    while( lex->lookAhead(0) == ',' ){
        lex->nextToken();
        QDomElement declarators;
        if( !parseInitDeclaratorList(declarators) ){
            lex->setIndex( index );
            return false;
        }
        decl.appendChild( declarators );
    }

    ADVANCE( ';', ';' );

    return true;
}

bool Parser::isConstructorDecl()
{
    if( lex->lookAhead(0) != '(' ){
        return false;
    }

    int tk = lex->lookAhead( 1 );
    switch( tk ){
    case '*':
    case '&':
    case '(':
        return false;

    case Token_const:
    case Token_volatile:
        return true;

    case Token_scope:
    {
        int index = lex->index();
        QDomElement temp;
        if( parsePtrToMember(temp) )
            return false;

        lex->setIndex( index );
    }
    break;

    }

    return true;
}

bool Parser::parseConstructorDeclaration( QDomElement& constr )
{
    if( lex->lookAhead(0) != '(' ){
        return false;
    }
    lex->nextToken();

    constr = m_dom->createElement( "constructor-declaration" );
    QDomElement args;

    if( !parseParameterDeclarationClause(args) ){
        return false;
    }
    constr.appendChild( args );

    if( lex->lookAhead(0) != ')' ){
        return false;
    }
    lex->nextToken();

    if( lex->lookAhead(0) == ':' ){
        QDomElement init;
        if( !parseCtorInitializer(init) ){
            return false;
        }
        constr.appendChild( init );
    }

    return true;
}

bool Parser::parseOperator( QDomElement& op )
{
    op = m_dom->createElement( "operator" );

    QString ide = lex->lookAhead( 0 ).toString();

    switch( lex->lookAhead(0) ){
    case Token_new:
    case Token_delete:
        lex->nextToken();
        if( lex->lookAhead(0) == '[' && lex->lookAhead(1) == ']' ){
            lex->nextToken();
            lex->nextToken();
            op.setAttribute( "id", ide + "[]" );
        } else {
            op.setAttribute( "id", ide );
        }
        return true;

    case '+':
    case '-':
    case '*':
    case '/':
    case '%':
    case '^':
    case '&':
    case '|':
    case '~':
    case '!':
    case '=':
    case '<':
    case '>':
    case ',':
    case Token_assign:
    case Token_shift:
    case Token_eq:
    case Token_not_eq:
    case Token_leq:
    case Token_geq:
    case Token_and:
    case Token_or:
    case Token_incr:
    case Token_decr:
    case Token_ptrmem:
    case Token_arrow:
        op.setAttribute( "id", ide );
        lex->nextToken();
        return true;

    default:
        if( lex->lookAhead(0) == '(' && lex->lookAhead(1) == ')' ){
            lex->nextToken();
            lex->nextToken();
            op.setAttribute( "id", "()" );
            return true;
        } else if( lex->lookAhead(0) == '[' && lex->lookAhead(1) == ']' ){
            lex->nextToken();
            lex->nextToken();
            op.setAttribute( "id", "[]" );
            return true;
        }
    }

    return false;
}

bool Parser::parseCvQualify( QDomElement& qualify )
{
    qualify = m_dom->createElement( "cv-qualify" );
    while( !lex->lookAhead(0).isNull() ){
        int tk = lex->lookAhead( 0 );
        if( tk == Token_const || tk == Token_volatile ){
            QDomElement e = m_dom->createElement( lex->lookAhead(0).toString() );
            qualify.appendChild( e );
            lex->nextToken();
        } else
            break;
    }
    return true; //qualify.hasChildNodes();
}

bool Parser::parseSimpleTypeSpecifier( QDomElement& spec )
{
    qDebug( "Parser::parseSimpleTypeSpecifier()" );

    spec = m_dom->createElement( "type-spec" );
    bool isIntegral = false;

    while( !lex->lookAhead(0).isNull() ){
        int tk = lex->lookAhead( 0 );

        if( tk == Token_char || tk == Token_wchar_t || tk == Token_bool || tk == Token_short ||
            tk == Token_int || tk == Token_long || tk == Token_signed || tk == Token_unsigned ||
            tk == Token_float || tk == Token_double || tk == Token_void ){

            isIntegral = true;
            spec.appendChild( m_dom->createElement( lex->lookAhead(0).toString() ) );
            lex->nextToken();
        } else if( isIntegral ){
            return true;
        } else
            break;
    }

    return parseName( spec );
}

bool Parser::parsePtrOperator( QDomElement& op )
{
    op = m_dom->createElement( "ptr-operator" );

    if( lex->lookAhead(0) == '&' ){
        op.setAttribute( "id", lex->lookAhead(0).toString() );
        lex->nextToken();
    } else if( lex->lookAhead(0) == '*' ){
        op.setAttribute( "id", lex->lookAhead(0).toString() );
        lex->nextToken();
    } else {
        int index = lex->index();
        if( !parsePtrToMember(op) ){
            lex->setIndex( index );
            return false;
        }
    }

    QDomElement cv;
    if( parseCvQualify(cv) ){
        op.appendChild( cv );
    }

    return true;
}


bool Parser::parseTemplateArgument( QDomElement& arg )
{
    if( parseTypeId(arg) ){
        return true;
    }

    return parseAssignmentExpression( arg );
}

bool Parser::parseTypeSpecifier( QDomElement& typeSpec )
{
    qDebug( "Parser::parseTypeSpecifier()" );

    QDomElement cv;
    parseCvQualify( cv );

    if( parseClassSpecifier(typeSpec) || parseEnumSpecifier(typeSpec) ||
        parseElaboratedTypeSpecifier(typeSpec) || parseSimpleTypeSpecifier(typeSpec) ){


        typeSpec.appendChild( cv );         // TODO: merge
        if( parseCvQualify(cv) ){
            typeSpec.appendChild( cv );
        }
    } else
        return false;

    return true;
}

bool Parser::parseDeclarator( QDomElement& declarator )
{
    declarator = m_dom->createElement( "declarator" );

    QDomElement ptrOp;
    while( parsePtrOperator(ptrOp) ){
        declarator.appendChild( ptrOp );
    }

    if( lex->lookAhead(0) == '(' ){
        lex->nextToken();

        QDomElement subDecl;
        if( !parseDeclarator(subDecl) ){
//             reportError( "expected a nested declarator",
//                          lex->lookAhead(0) );
            return false;
        }
        declarator.appendChild( subDecl );
        if( lex->lookAhead(0) != ')'){
            return false;
        }
        lex->nextToken();
    } else {
        QDomElement id;
        if( !parseDeclaratorId(id) ){
            return false;
        }
        declarator.appendChild( id );

        if( lex->lookAhead(0) == ':' ){
            declarator.setAttribute( "bit-declaration", true );
            lex->nextToken();
            QDomElement initExpr;
            if( !parseConstantExpression(initExpr) ){
                return false;
            }
            declarator.appendChild( initExpr );
            return true;
        }

    }

    while( lex->lookAhead(0) == '[' ){
        QDomElement arr = m_dom->createElement( "array" );
        lex->nextToken();
        QDomElement expr;
        if( !parseCommaExpression(expr) ){
//             reportError( "expected comma expression",
//                          lex->lookAhead(0) );
            return false;
        }
        arr.appendChild( expr );
        declarator.appendChild( arr );
        if( lex->lookAhead(0) != ']' ){
            return false;
        }
        lex->nextToken();
    }

    if( lex->lookAhead(0) == '(' ){
        // parse function arguments
        lex->nextToken();

        QDomElement parameterDeclarationClause;
        if( !parseParameterDeclarationClause(parameterDeclarationClause) ){
            return false;
        }
        declarator.appendChild( parameterDeclarationClause );
        if( lex->lookAhead(0) != ')' ){
            return false;
        }
        lex->nextToken();

        QDomElement cv;
        if( parseCvQualify(cv) ){
            declarator.appendChild( cv );
        }

        QDomElement exceptionSpecification;
        if( parseExceptionSpecification(exceptionSpecification) ){
            declarator.appendChild( exceptionSpecification );
        }
    }

    return true;
}

bool Parser::parseEnumSpecifier( QDomElement& spec )
{
    int index = lex->index();

    if( lex->lookAhead(0) != Token_enum ){
        return false;
    }

    spec = m_dom->createElement( "enum-specifier" );
    lex->nextToken();

    if( lex->lookAhead(0) == Token_identifier ){
        spec.setAttribute( "id", lex->lookAhead(0).toString() );
        lex->nextToken();
    }

    if( lex->lookAhead(0) != '{' ){
        lex->setIndex( index );
        return false;
    }

    ADVANCE( '{', '{' );

    QDomElement enumeratorList;
    if( parseEnumeratorList( enumeratorList ) ){
        spec.appendChild( enumeratorList );
    }

    ADVANCE( '}', '}' );

    return true;
}

bool Parser::parseTemplateParameterList( QDomElement& parameterList )
{
    parameterList = m_dom->createElement( "parameter-list" );

    QDomElement param;
    if( !parseTemplateParameter(param) ){
        return false;
    }
    parameterList.appendChild( param );

    while( lex->lookAhead(0) == ',' ){
        lex->nextToken();

        if( parseTemplateParameter(param) ){
            parameterList.appendChild( param );
        } else {
            // error recovery
            return false;
        }
    }

    return true;
}

bool Parser::parseTemplateParameter( QDomElement& param )
{
    int tk = lex->lookAhead( 0 );
    if( tk == Token_class ||
        tk == Token_typename ||
        tk == Token_template )
        return parseTypeParameter( param );
    else
        return parseParameterDeclaration( param );
}

bool Parser::parseTypeParameter( QDomElement& param )
{
    switch( lex->lookAhead(0) ){

    case Token_class:
    {
        param = m_dom->createElement( "class-parameter" );

        lex->nextToken(); // skip class

        // parse optional name
        QString name;
        if( lex->lookAhead(0) == Token_identifier ){
            name = lex->lookAhead( 0 ).toString();
            lex->nextToken();
        }
        if( name )
            param.setAttribute( "id", name );

        if( lex->lookAhead(0) == '=' ){
            lex->nextToken();

            QDomElement typeId;
            if( !parseTypeId(typeId) ){
                reportError( "expected a type id", lex->lookAhead(0) );
                return false;
            }

            param.appendChild( typeId );
        }
    }
    return true;

    case Token_typename:
    {
        param = m_dom->createElement( "typename-parameter" );

        lex->nextToken(); // skip typename

        // parse optional name
        QString name;
        if( lex->lookAhead(0) == Token_identifier ){
            name = lex->lookAhead( 0 ).toString();
            lex->nextToken();
        }
        if( name )
            param.setAttribute( "id", name );

        if( lex->lookAhead(0) == '=' ){
            lex->nextToken();

            QDomElement typeId;
            if( !parseTypeId(typeId) ){
                reportError( "expected a type id", lex->lookAhead(0) );
                return false;
            }

            param.appendChild( typeId );
        }
    }
    return true;

    case Token_template:
    {
        param = m_dom->createElement( "template-parameter" );

        lex->nextToken(); // skip template
        ADVANCE( '<', '<' );
        QDomElement parameterList;
        if( !parseTemplateParameterList(parameterList) ){
            return false;
        }
        param.appendChild( parameterList );

        ADVANCE( '>', '>' );
        ADVANCE( Token_class, "class" );

        // parse optional name
        QString name;
        if( lex->lookAhead(0) == Token_identifier ){
            name = lex->lookAhead( 0 ).toString();
            lex->nextToken();
        }
        if( name )
            param.setAttribute( "id", name );

        if( lex->lookAhead(0) == '=' ){
            lex->nextToken();

            QString templ_name = lex->lookAhead( 0 ).toString();
            if( lex->lookAhead(0) != Token_identifier ){
                reportError( "expected a template name", lex->lookAhead(0) );
                return false;
            }
            param.setAttribute( "default", templ_name );

            lex->nextToken(); // skip template-name
        }
    }
    return true;

    } // end switch


    return false;
}

bool Parser::parseStorageClassSpecifier( QDomElement& spec )
{
    switch( lex->lookAhead(0) ){
    case Token_friend:
    case Token_auto:
    case Token_register:
    case Token_static:
    case Token_extern:
    case Token_mutable:
        spec = m_dom->createElement( lex->lookAhead(0).toString() );
        lex->nextToken();
        return true;
    }

    return false;
}

bool Parser::parseFunctionSpecifier( QDomElement& spec )
{
    switch( lex->lookAhead(0) ){
    case Token_inline:
    case Token_virtual:
    case Token_explicit:
        spec = m_dom->createElement( lex->lookAhead(0).toString() );
        lex->nextToken();
        return true;
    }

    return false;
}

// type-id:
//         type-specifier-seq abstract-declaratoropt
bool Parser::parseTypeId( QDomElement& id )
{
    id = m_dom->createElement( "type-id" );

    QDomElement typeSpec;
    if( !parseTypeSpecifier(typeSpec) ){
        return false;
    }
    id.appendChild( typeSpec );

    QDomElement declarator;
    if( parseAbstractDeclarator(declarator) ){
        id.appendChild( declarator );
    }

    return true;
}

bool Parser::parseAbstractDeclarator( QDomElement& declarator )
{
#warning "TODO Parser::parseAbstractDeclarator()"
#warning "resolve ambiguities with sub-abstract-declarator and function argument"

    declarator = m_dom->createElement( "abstract-declarator" );

    QDomElement ptrOp;
    while( parsePtrOperator(ptrOp) ){
        declarator.appendChild( ptrOp );
    }

    if( lex->lookAhead(0) == '(' ){
        lex->nextToken();

        QDomElement subDecl;
        if( !parseDeclarator(subDecl) ){
//             reportError( "expected a nested declarator",
//                          lex->lookAhead(0) );
            return false;
        }
        declarator.appendChild( subDecl );
        if( lex->lookAhead(0) != ')'){
            return false;
        }
        lex->nextToken();
    }

    while( lex->lookAhead(0) == '[' ){
        QDomElement arr = m_dom->createElement( "array" );
        lex->nextToken();
        QDomElement expr;
        if( !parseCommaExpression(expr) ){
//             reportError( "expected comma expression",
//                          lex->lookAhead(0) );
            return false;
        }
        arr.appendChild( expr );
        declarator.appendChild( arr );
        if( lex->lookAhead(0) != ']' ){
            return false;
        }
        lex->nextToken();
    }

    if( lex->lookAhead(0) == '(' ){
        // parse function arguments
        lex->nextToken();

        QDomElement parameterDeclarationClause;
        if( !parseParameterDeclarationClause(parameterDeclarationClause) ){
            return false;
        }
        declarator.appendChild( parameterDeclarationClause );
        if( lex->lookAhead(0) != ')' ){
            return false;
        }
        lex->nextToken();

        QDomElement cv;
        if( parseCvQualify(cv) ){
            declarator.appendChild( cv );
        }

        QDomElement exceptionSpecification;
        if( parseExceptionSpecification(exceptionSpecification) ){
            declarator.appendChild( exceptionSpecification );
        }
    }

    return true;
}

bool Parser::parseConstantExpression( QDomElement& expr )
{
    expr = m_dom->createElement( "constant-expression" );

    QStringList l;

    while( !lex->lookAhead(0).isNull() ){
        int tk = lex->lookAhead( 0 );

        if( tk == '(' ){
            l << "(...)";
            if( !skip('(', ')') ){
                return false;
            }
            lex->nextToken();
        } else if( tk == ',' || tk == ';' || tk == '<'
            || tk == Token_assign || tk == ']' || tk == ')' || tk == '}' ){
            break;
        } else {
            l << lex->lookAhead( 0 ).toString();
            lex->nextToken();
        }
    }

    if( !l.isEmpty() ){
        expr.setAttribute( "body", l.join(" ") );
    }

    return true;
}


bool Parser::parseInitDeclaratorList( QDomElement& declaratorList )
{
    declaratorList = m_dom->createElement( "init-declarator-list" );

    QDomElement initDecl;
    if( !parseInitDeclarator(initDecl) ){
        return false;
    }
    declaratorList.appendChild( initDecl );

    while( lex->lookAhead(0) == ',' ){
        lex->nextToken();

        if( parseInitDeclarator(initDecl) ){
            declaratorList.appendChild( initDecl );
        } else
            break;
    }
    return true;
}

bool Parser::parseFunctionBody( QDomElement& body )
{
    body = m_dom->createElement( "function-body" );

    if( lex->lookAhead(0) != '{' ){
        return false;
    }

    if( skip('{', '}') ){
        ADVANCE( '}', '}' );
    } else
        return false;

    return true;
}

bool Parser::parseParameterDeclarationClause( QDomElement& clause )
{
    clause = m_dom->createElement( "parameter-declaration-clause" );

    QDomElement declList;
    if( parseParameterDeclarationList(declList) ){
        clause.appendChild( declList );
    }

    if( lex->lookAhead(0) == ',' ){
        lex->nextToken();
    }

    if( lex->lookAhead(0) == Token_ellipsis ){
        clause.setAttribute( "ellipsis", true );
        lex->nextToken();
    }

    return true;
}

bool Parser::parseParameterDeclarationList( QDomElement& declList)
{
    declList = m_dom->createElement( "parameter-declaration-list" );

    QDomElement decl;
    if( !parseParameterDeclaration(decl) ){
        return false;
    }
    declList.appendChild( decl );

    while( lex->lookAhead(0) == ',' ){
        lex->nextToken();

        if( parseParameterDeclaration(decl) ){
            declList.appendChild( decl );
        } else
            break;
    }
    return true;
}

bool Parser::parseParameterDeclaration( QDomElement& param )
{
    param = m_dom->createElement( "parameter-declaration" );

    // parse decl spec
    QDomElement typeSpec;
    if( !parseTypeSpecifier(typeSpec) ){
        return false;
    }
    param.appendChild( typeSpec );

    int index = lex->index();
    QDomElement declarator;
    if( parseDeclarator(declarator) ){
        param.appendChild( declarator );
    } else {
        lex->setIndex( index );
        // optional abstract declarator
        if( parseAbstractDeclarator(declarator) ){
            param.appendChild( declarator );
        }
    }

    if( lex->lookAhead(0) == '=' ){
        lex->nextToken();
        QDomElement expr;
        if( !parseAssignmentExpression(expr) ){
            return false;
        }
        param.appendChild( expr );
    }

    return true;
}

bool Parser::parseClassSpecifier( QDomElement& classSpec )
{
    int index = lex->index();

    if( !parseClassHead(classSpec) ){
        return false;
    }

    if( lex->lookAhead(0) != '{' ){
        lex->setIndex( index );
        return false;
    }

    ADVANCE( '{', '{' );

    if( lex->lookAhead(0) != '}' ){
        QDomElement memberSpecificationList;
        if( parseMemberSpecificationList(memberSpecificationList) ){
            classSpec.appendChild( memberSpecificationList );
        }
    }

    ADVANCE( '}', '}' );

    return true;
}

bool Parser::parseAccessSpecifier( QDomElement& spec )
{
    switch( lex->lookAhead(0) ){
    case Token_public:
    case Token_protected:
    case Token_private:
        spec = m_dom->createElement( lex->lookAhead(0).toString() );
        lex->nextToken();
        return true;
    }

    return false;
}

bool Parser::parseClassHead( QDomElement& spec )
{
    int tk = lex->lookAhead( 0 );
    if( tk == Token_class || tk == Token_struct || tk == Token_union ){
        spec = m_dom->createElement( lex->lookAhead(0).toString() + "-specifier" );
        lex->nextToken();
    } else {
        return false;
    }


    QDomElement name;
    if( parseName(name) ){
        spec.appendChild( name );
    }

    QDomElement baseClause;
    if( parseBaseClause(baseClause) ){
        spec.appendChild( baseClause );
    }

    return true;
}

bool Parser::parseMemberSpecificationList( QDomElement& memberSpecificationList )
{
    memberSpecificationList = m_dom->createElement( "member-specification-list" );

    QDomElement memberSpecification;
    if( !parseMemberSpecification(memberSpecification) ){
        return false;
    }
    memberSpecificationList.appendChild( memberSpecification );

    while( lex->lookAhead(0) != '}' ){
        if( parseMemberSpecification(memberSpecification) ){
            memberSpecificationList.appendChild( memberSpecification );
        } else
            break;
    }

    return true;
}

bool Parser::parseMemberSpecification( QDomElement& memberSpecification )
{
    if( lex->lookAhead(0) == ';' ){
        memberSpecification = m_dom->createElement( "null-declaration" );
        lex->nextToken();
        return true;
    } else if( lex->lookAhead(0) == Token_Q_OBJECT || lex->lookAhead(0) == Token_K_DCOP ){
        memberSpecification = m_dom->createElement( lex->lookAhead(0).toString() );
        lex->nextToken();
        return true;
    } else if( lex->lookAhead(0) == Token_signals || lex->lookAhead(0) == Token_k_dcop ){
        memberSpecification = m_dom->createElement( lex->lookAhead(0).toString() );
        lex->nextToken();
        ADVANCE( ':', ':' );
        return true;
    } else if( parseTypedef(memberSpecification) ){
        return true;
    } else if( parseUsing(memberSpecification) ){
        return true;
    } else if( parseTemplateDeclaration(memberSpecification) ){
        return true;
    } else if( parseAccessSpecifier(memberSpecification) ){
        if( lex->lookAhead(0) == Token_slots ){
            memberSpecification.appendChild( m_dom->createElement("slots") );
            lex->nextToken();
        }
        ADVANCE( ':', ':' );
        return true;
    }

    return parseDeclaration( memberSpecification );
}

bool Parser::parseCtorInitializer( QDomElement& ctor )
{
    ctor = m_dom->createElement( "ctor-initializer" );

    if( lex->lookAhead(0) != ':' ){
        return false;
    }
    lex->nextToken();

    QDomElement memInitializerList;
    if( !parseMemInitializerList(memInitializerList) ){
        reportError( "expected member initializer list", lex->lookAhead(0) );
        return false;
    }
    ctor.appendChild( memInitializerList );

    return true;
}

// elaborated-type-specifier:
//         class-key ::opt nested-name-specifieropt identifier
//         enum ::opt nested-name-specifieropt identifier
//         typename ::opt nested-name-specifier identifier
//         typename ::opt nested-name-specifier identifier < template-argument-list >
bool Parser::parseElaboratedTypeSpecifier( QDomElement& spec )
{
    spec = m_dom->createElement( "elaborated-type-spec" );

    int index = lex->index();

    int tk = lex->lookAhead( 0 );
    if( tk == Token_class || tk == Token_struct || tk == Token_union ){
        spec.setAttribute( "type", lex->lookAhead(0).toString() );
        lex->nextToken();

        QDomElement name;
        if( parseName(name) ){
            spec.appendChild( name );
            return true;
        }
    } else if( tk == Token_enum ){
        spec.setAttribute( "type", lex->lookAhead(0).toString() );
        lex->nextToken();

        QDomElement name;
        if( parseName(name) ){
            spec.appendChild( name );
            return true;
        }
    } else if( tk == Token_typename ){
        spec.setAttribute( "type", lex->lookAhead(0).toString() );
        lex->nextToken();

        QDomElement name;
        if( parseName(name) ){
            spec.appendChild( name );
            return true;
        }
    }

    lex->setIndex( index );
    return false;
}

bool Parser::parseDeclaratorId( QDomElement& id )
{
    return parseName( id );
}

bool Parser::parseCommaExpression( QDomElement& expr )
{
#warning "TODO Parser::parseCommaExpression()"

    expr = m_dom->createElement( "comma-expression" );
    while( !lex->lookAhead(0).isNull() ){
        int tk = lex->lookAhead( 0 );
        if( tk == '(' ){
            if( !skip('(',')') ){
                return false;
            }
        } else if( tk == ']' || tk == ';' ){
            break;
        }
        lex->nextToken();
    }
    return true;
}

bool Parser::parseExceptionSpecification( QDomElement& spec )
{
    if( lex->lookAhead(0) != Token_throw ){
        return false;
    }
    lex->nextToken();

    spec = m_dom->createElement( "exception-specification" );

    ADVANCE( '(', '(' );
    QDomElement typeIdList;
    if( !parseTypeIdList(typeIdList) ){
        reportError( "expected type id list", lex->lookAhead(0) );
        return false;
    }
    spec.appendChild( typeIdList );
    ADVANCE( ')', ')' );

    return true;
}

bool Parser::parseEnumeratorList( QDomElement& enumeratorList )
{
    enumeratorList = m_dom->createElement( "enumerator-list" );

    QDomElement enumerator;
    if( !parseEnumerator(enumerator) ){
        return false;
    }
    enumeratorList.appendChild( enumerator );

    while( lex->lookAhead(0) == ',' ){
        lex->nextToken();

        QDomElement enumerator;
        if( parseEnumerator(enumerator) ){
            enumeratorList.appendChild( enumerator );
        } else
            break;
    }

    return true;
}

bool Parser::parseEnumerator( QDomElement& enumerator )
{
    enumerator = m_dom->createElement( "enumerator" );

    if( lex->lookAhead(0) != Token_identifier ){
        return false;
    }

    QString ide = lex->lookAhead( 0 ).toString();
    enumerator.setAttribute( "id", ide );
    lex->nextToken();

    if( lex->lookAhead(0) == '=' ){
        lex->nextToken();

        QDomElement expr;
        if( !parseConstantExpression(expr) ){
            return false;
        }
        enumerator.appendChild( expr );
    }
    return true;
}

bool Parser::parseInitDeclarator( QDomElement& initDecl )
{
    qDebug( "Parser::parseInitDeclarator()" );
    initDecl = m_dom->createElement( "init-declarator" );

    QDomElement declarator;
    if( !parseDeclarator(declarator) ){
        return false;
    }
    initDecl.appendChild( declarator );

    QDomElement initializer;
    if( parseInitializer(initializer) ){
        initDecl.appendChild( initDecl );
    }
    initDecl.appendChild( initializer );

    return true;
}

bool Parser::parseAssignmentExpression( QDomElement& expr )
{
#warning "TODO Parser::parseAssignmentExpression()"

    expr = m_dom->createElement( "assignment-expression" );

    QStringList l;

    while( !lex->lookAhead(0).isNull() ){
        int tk = lex->lookAhead( 0 );

        if( tk == '(' ){
            if( !skip('(', ')') ){
                return false;
            }
        } else if( tk == ',' || tk == ';' || tk == '<' || tk == '>' ||
                   tk == Token_assign || tk == ']' || tk == ')' ){
            break;
        }

        l << lex->lookAhead( 0 ).toString();
        lex->nextToken();
    }

    if( !l.isEmpty() ){
        expr.setAttribute( "body", l.join(" ") );
    }

    return true;
}

bool Parser::parseBaseClause( QDomElement& baseClause )
{
    if( lex->lookAhead(0) != ':' ){
        return false;
    }
    lex->nextToken();

    baseClause = m_dom->createElement( "base-clause" );

    QDomElement baseSpecifierList;
    if( !parseBaseSpecifierList(baseSpecifierList) ){
        reportError( "expected base specifier list", lex->lookAhead(0) );
        return false;
    }
    baseClause.appendChild( baseSpecifierList );

    return true;
}

bool Parser::parseInitializer( QDomElement& init )
{
    init = m_dom->createElement( "initializer" );

    qDebug( "Parser::parseInitializer()" );

    if( lex->lookAhead(0) == '=' ){
        lex->nextToken();

        QDomElement clause;
        if( !parseInitializerClause(clause) ){
            reportError( "expected initializer clause", lex->lookAhead(0) );
            return false;
        }
        init.appendChild( clause );
    } else if( lex->lookAhead(0) == '(' ){
        lex->nextToken();
        QDomElement expr;
        if( !parseExpressionList(expr) ){
            reportError( "expected expression list", lex->lookAhead(0) );
            return false;
        }
        init.appendChild( expr );
        ADVANCE( ')', ')' );
    } else
        return false;

    return true;
}

bool Parser::parseMemInitializerList( QDomElement& memInitializerList )
{
    memInitializerList = m_dom->createElement( "mem-initializer-list" );

    QDomElement memInitializer;
    if( !parseMemInitializer(memInitializer) ){
        return false;
    }
    memInitializerList.appendChild( memInitializer );

    while( lex->lookAhead(0) == ',' ){
        lex->nextToken();

        if( parseMemInitializer(memInitializer) ){
            memInitializerList.appendChild( memInitializer );
        } else
            break;
    }

    return true;
}

bool Parser::parseMemInitializer( QDomElement& memInitializer )
{
    memInitializer = m_dom->createElement( "mem-initializer" );

    QDomElement id;
    if( !parseMemInitializerId(id) ){
        reportError( "expected member initializer", lex->lookAhead(0) );
        return false;
    }
    memInitializer.appendChild( id );
    ADVANCE( '(', '(' );
    QDomElement exprList;
    if( !parseExpressionList(exprList) ){
        reportError( "expected an expression list", lex->lookAhead(0) );
        return false;
    }
    ADVANCE( ')', ')' );
    return true;
}

bool Parser::parseTypeIdList( QDomElement& typeIdList )
{
    typeIdList = m_dom->createElement( "type-id-list" );

    QDomElement typeId;
    if( !parseTypeId(typeId) ){
        return false;
    }

    while( lex->lookAhead(0) == ',' ){
        if( parseTypeId(typeId) ){
            typeIdList.appendChild( typeId );
        } else
            break;
    }

    return true;
}

bool Parser::parseBaseSpecifierList( QDomElement& baseSpecifierList)
{
    baseSpecifierList = m_dom->createElement( "base-specifier-list" );

    QDomElement baseSpecifier;
    if( !parseBaseSpecifier(baseSpecifier) ){
        return false;
    }
    baseSpecifierList.appendChild( baseSpecifier );

    while( lex->lookAhead(0) == ',' ){
        lex->nextToken();

        if( parseBaseSpecifier(baseSpecifier) ){
            baseSpecifierList.appendChild( baseSpecifier );
        } else
            break;
    }

    return true;
}

bool Parser::parseBaseSpecifier( QDomElement& baseSpecifier )
{
    baseSpecifier = m_dom->createElement( "base-specifier" );

    if( lex->lookAhead(0) == Token_virtual ){
        baseSpecifier.setAttribute( "virtual", true );
        lex->nextToken();

        QDomElement accessSpec;
        if( parseAccessSpecifier(accessSpec) ){
            baseSpecifier.appendChild( accessSpec );
        }
    } else {

        QDomElement accessSpec;
        if( parseAccessSpecifier(accessSpec) ){
            baseSpecifier.appendChild( accessSpec );
        }

        if( lex->lookAhead(0) == Token_virtual ){
            baseSpecifier.setAttribute( "virtual", true );
            lex->nextToken();
        }
    }

    if( lex->lookAhead(0) == Token_scope ){
        lex->nextToken();
    }

    QDomElement name;
    if( !parseName(name) ){
        reportError( "expected class name", lex->lookAhead(0) );
        return false;
    }
    baseSpecifier.appendChild( name );

    return true;
}


bool Parser::parseInitializerClause( QDomElement& clause )
{
#warning "TODO initializer-list"

    clause = m_dom->createElement( "initializer-clause" );

    if( lex->lookAhead(0) == '{' ){
        if( skip('{','}') ){
            ADVANCE( '}', '}' );
        } else
            return false;
    } else {
        QDomElement expr;
        if( !parseAssignmentExpression(expr) ){
            reportError( "expected assignment expected", lex->lookAhead(0) );
            return false;
        }
        clause.appendChild( expr );
    }

    return true;
}

bool Parser::parseMemInitializerId( QDomElement& id )
{
    return parseName( id );
}


bool Parser::parseExpressionList( QDomElement& exprList )
{
    exprList = m_dom->createElement( "expression-list" );
    while( !lex->lookAhead(0).isNull() ){
        int tk = lex->lookAhead( 0 );
        if( tk == '(' ){
            if( !skip('(', ')') ){
                return false;
            }
            ADVANCE( ')', ')' );
        } else if( tk == ')' ){
            break;
        } else
            lex->nextToken();
    }
    return true;
}

bool Parser::parseNestedNameSpecifier( QDomElement& name )
{
    name = m_dom->createElement( "nested-name-specifier" );

    if( lex->lookAhead(0) == Token_scope ){
        name.setAttribute( "global", true );
        lex->nextToken();
    }

    while( lex->lookAhead(0) == Token_identifier ){
        name.appendChild( m_dom->createElement(lex->lookAhead(0).toString()) );
        lex->nextToken();

        if( lex->lookAhead(0) == Token_scope ){
            lex->nextToken();
        } else
            break;
    }

    return true;
}

bool Parser::parsePtrToMember( QDomElement& name )
{
    name = m_dom->createElement( "ptr-to-member" );

    if( lex->lookAhead(0) == Token_scope ){
        name.setAttribute( "global", true );
        lex->nextToken();
    }

    while( lex->lookAhead(0) == Token_identifier ){
        name.appendChild( m_dom->createElement(lex->lookAhead(0).toString()) );
        lex->nextToken();

        if( lex->lookAhead(0) == Token_scope && lex->lookAhead(1) == '*' ){
            lex->nextToken(); // skip scope
            lex->nextToken(); // skip *
            return true;
        } else
            break;
    }

    return false;
}
